import React, { forwardRef, useEffect, useMemo, useRef, useState } from 'react';
import debounce from 'lodash.debounce';
import * as monaco from 'monaco-editor';
import { createEditor, getMonacoLanguage, updateEditorKeybindingsMode, registerEditorOpenHandler, createModelReference, registerFile } from '@codingame/monaco-editor-wrapper';
import { DisposableStore } from 'vscode/monaco';
import { RegisteredMemoryFile } from '@codingame/monaco-vscode-files-service-override';
import { useDeepMemo, useLastValueRef, useLastVersion, useThemeColor } from './hooks.js';
import './style.js';
const STATUS_BAR_HEIGHT = 20;
function fixCode(code) {
    // If monaco editor is provided code using \r\n
    // The editor seems to flicker and the language server is spammed by document/didChange events
    // TODO investigate me?
    return code.replace(/\r\n?/g, '\n');
}
// Some editor properties are managed directly by monaco-vim / monaco-emacs so we don't want to override them
const keyBindingsManagedOptions = {
    vim: new Set(['cursorBlinking', 'cursorWidth']),
    emacs: new Set(['cursorBlinking', 'cursorStyle']),
    classic: null
};
function removeKeyBindingsManagedOptions(options, keyBindingsMode) {
    const managedOptions = keyBindingsManagedOptions[keyBindingsMode];
    if (managedOptions == null) {
        return options;
    }
    else {
        return Object.fromEntries(Object.entries(options)
            .filter(([optionId]) => !managedOptions.has(optionId)));
    }
}
const viewStates = new Map();
export function defaultRestoreViewState(editor, model) {
    const viewState = viewStates.get(model.uri.toString());
    if (viewState != null) {
        editor.restoreViewState(viewState);
    }
}
export function defaultSaveViewState(editor, model) {
    const viewState = editor.saveViewState();
    const key = model.uri.toString();
    if (viewState != null) {
        viewStates.set(key, viewState);
    }
    else {
        viewStates.delete(key);
    }
}
function defaultOnDidSave() {
}
function MonacoEditor({ height: requestedHeight = '100%', maxHeight, minHeight = 350, fileUri, programmingLanguage, value, options, onChange, keyBindingsMode = 'classic', markers, saveViewState = defaultSaveViewState, restoreViewState = defaultRestoreViewState, onEditorOpenRequest, onDidSave = defaultOnDidSave }, ref) {
    const editorRef = useRef();
    const modelRef = useRef();
    const [modelReady, setModelReady] = useState(false);
    const preventTriggerChangeEventRef = useRef(false);
    const [height, setHeight] = useState(requestedHeight !== 'auto' ? requestedHeight : 50);
    const statusBarBackground = useThemeColor('statusBar.background');
    const statusBarForeground = useThemeColor('statusBar.foreground');
    const statusBarBorder = useThemeColor('statusBar.border');
    const containerRef = useRef(null);
    const statusBarRef = useRef(null);
    const monacoLanguage = useMemo(() => programmingLanguage != null ? getMonacoLanguage(programmingLanguage) : undefined, [programmingLanguage]);
    const memoizedOptions = useDeepMemo(() => options, [options]);
    const allOptions = useMemo(() => {
        return removeKeyBindingsManagedOptions({
            ...memoizedOptions
        }, keyBindingsMode);
    }, [memoizedOptions, keyBindingsMode]);
    const fixedCode = useMemo(() => value != null ? fixCode(value) : null, [value]);
    const valueRef = useLastValueRef(fixedCode);
    const lastSaveViewState = useLastVersion(saveViewState);
    const lastRestoreViewState = useLastVersion(restoreViewState);
    const lastOnDidSave = useLastVersion(onDidSave);
    const hasValue = fixedCode != null;
    // Create editor
    useEffect(() => {
        const containerElement = containerRef.current;
        if (containerElement != null) {
            const editor = createEditor(containerElement, {
                automaticLayout: true,
                // We need to pass options here for the `scrollbar` options to be used
                ...allOptions,
                // We need to override all IStandaloneEditorConstructionOptions fields to prevent conflicts with proper editor options (especially `language`)
                value: undefined,
                language: undefined,
                theme: undefined,
                model: null,
                autoDetectHighContrast: undefined,
                accessibilityHelpUrl: undefined,
                ariaContainerElement: undefined,
                dimension: undefined,
                overflowWidgetsDomNode: undefined
            });
            editorRef.current = editor;
            if (ref != null) {
                if (typeof ref === 'function') {
                    ref(editor);
                }
                else {
                    ref.current = editor;
                }
            }
            return () => {
                editorRef.current = undefined;
                editor.dispose();
            };
        }
        return undefined;
        // eslint-disable-next-line react-hooks/exhaustive-deps
    }, []);
    // Create/Update model
    useEffect(() => {
        if (fileUri == null && !hasValue) {
            modelRef.current = undefined;
            editorRef.current.setModel(null);
            return;
        }
        const disposableStore = new DisposableStore();
        async function updateModel() {
            modelRef.current = undefined;
            editorRef.current.setModel(null);
            setModelReady(false);
            const value = valueRef.current;
            let modelIRef;
            let model;
            if (fileUri != null) {
                const uri = monaco.Uri.parse(fileUri);
                disposableStore.add(registerFile(new RegisteredMemoryFile(uri, value)));
                const modelIRefPromise = createModelReference(uri);
                disposableStore.add({
                    dispose() {
                        void modelIRefPromise.then(modelIRef => modelIRef.dispose(), console.error);
                    }
                });
                modelIRef = (await modelIRefPromise);
                if (disposableStore.isDisposed) {
                    return;
                }
                disposableStore.add(modelIRef.object.onDidSave(lastOnDidSave));
                model = modelIRef.object.textEditorModel;
                if (monacoLanguage != null && model.getLanguageId() !== monacoLanguage) {
                    monaco.editor.setModelLanguage(model, monacoLanguage);
                }
            }
            else {
                model = monaco.editor.createModel(value, monacoLanguage);
                disposableStore.add({
                    dispose() {
                        model.dispose();
                    }
                });
            }
            modelRef.current = model;
            setModelReady(true);
            editorRef.current.setModel(model);
            if (editorRef.current != null) {
                lastRestoreViewState(editorRef.current, model);
            }
            disposableStore.add({
                dispose() {
                    if (editorRef.current != null) {
                        lastSaveViewState(editorRef.current, model);
                    }
                }
            });
        }
        updateModel().catch(console.error);
        return () => {
            disposableStore.dispose();
        };
    }, [monacoLanguage, fileUri, valueRef, lastSaveViewState, lastRestoreViewState, hasValue, lastOnDidSave]);
    // Update value
    useEffect(() => {
        if (modelReady && fixedCode != null) {
            const model = modelRef.current;
            const editor = editorRef.current;
            if (model != null && fixedCode !== model.getValue()) {
                preventTriggerChangeEventRef.current = true;
                console.debug('Replacing whole editor content');
                editor.pushUndoStop();
                model.setValue(fixedCode);
                editor.pushUndoStop();
                preventTriggerChangeEventRef.current = false;
            }
        }
    }, [fixedCode, modelReady]);
    // Update options from props
    useEffect(() => {
        editorRef.current.updateOptions(allOptions);
    }, [allOptions]);
    // Keybindings mode
    useEffect(() => {
        const editor = editorRef.current;
        const disposable = updateEditorKeybindingsMode(editor, keyBindingsMode, statusBarRef.current);
        return () => {
            disposable.dispose();
        };
    }, [keyBindingsMode]);
    // Update markers
    useEffect(() => {
        if (!modelReady) {
            return;
        }
        const model = modelRef.current;
        if (markers != null && model != null) {
            monaco.editor.setModelMarkers(model, 'customMarkers', markers);
            return () => {
                monaco.editor.setModelMarkers(model, 'customMarkers', []);
            };
        }
        return undefined;
    }, [markers, modelReady]);
    // Call onChange callback
    useEffect(() => {
        if (onChange != null) {
            // Sometimes, monaco calls the didChangeModelContent multiple times in a row
            // (For instance, after an autocomplete with some additionalTextEdits - used by the typescript language server)
            // The onChange prop is very probably used to store the current code in a state somewhere
            // Updating that state 2 times in a row will trigger 2 asynchronous renders
            // The first render will use the state after the first model content change, while the monaco model is already up to date with the last change
            // It will then produce a desync and this component will overwrite the editor content to sync it back
            // Doing so produces at least 2 issues: the cursor is moved to the beginning of the file by monaco and the undo stack is lost
            // So a solution is to debounce the onChange callback so it's called only once in that case
            const debouncedOnChange = debounce(onChange, 0);
            const editor = editorRef.current;
            const didChangeModelContentDisposable = editor.onDidChangeModelContent(event => {
                if (!preventTriggerChangeEventRef.current) {
                    debouncedOnChange(editor.getValue(), event);
                }
            });
            return () => {
                didChangeModelContentDisposable.dispose();
            };
        }
        return undefined;
    }, [onChange]);
    useEffect(() => {
        if (onEditorOpenRequest != null) {
            const disposable = registerEditorOpenHandler(async (model, input, source, sideBySide) => {
                if (source === editorRef.current) {
                    return onEditorOpenRequest(model, input, source, sideBySide);
                }
                return null;
            });
            return () => {
                disposable.dispose();
            };
        }
        return undefined;
    }, [onEditorOpenRequest]);
    // Compute height
    useEffect(() => {
        if (requestedHeight !== 'auto') {
            setHeight(requestedHeight);
            return;
        }
        const editor = editorRef.current;
        const recomputeHeight = () => {
            let height = Math.max(minHeight, editor.getContentHeight() + STATUS_BAR_HEIGHT);
            if (maxHeight != null) {
                height = Math.min(height, maxHeight);
            }
            setHeight(height);
        };
        const onDidContentSizeChangeDisposable = editor.onDidContentSizeChange(debounce(recomputeHeight, 50));
        recomputeHeight();
        return () => {
            onDidContentSizeChangeDisposable.dispose();
        };
    }, [minHeight, maxHeight, requestedHeight]);
    const statusBarStyle = useMemo(() => {
        return {
            backgroundColor: statusBarBackground ?? '#007ACC',
            color: statusBarForeground ?? '#FFFFFF',
            borderTop: `1px solid ${statusBarBorder ?? '#FFFFFF'}`
        };
    }, [statusBarBackground, statusBarBorder, statusBarForeground]);
    return (React.createElement("div", { className: 'react-monaco-editor-react-container', style: { height } },
        React.createElement("div", { className: 'react-monaco-editor-react-wrapper' },
            React.createElement("div", { className: 'react-monaco-editor-react', ref: containerRef }),
            React.createElement("div", { className: 'react-monaco-editor-react-status-bar', ref: statusBarRef, style: statusBarStyle }))));
}
export default forwardRef(MonacoEditor);
//# sourceMappingURL=MonacoEditor.js.map